// 06/10/09 added 64-bit stuff; part of common-1.0
// 10/08/08 add detection for Intel compiler (it rocks!)

#ifndef PLATFORM_HPP
#define PLATFORM_HPP

namespace cat {


//// Compiler ////

// Structure packing syntax
#if defined(__INTEL_COMPILER) || defined(__ICL) || defined(__ICC) || defined(__ECC)
# define COMPILER_ICC /* Intel C++ Compiler */
# define COMPILER_MSVC /* Compatible with MSVC */

# define PACKED __declspec(align(1))
# define INLINE __inline
# define DEBUG _DEBUG
            }
# include <cstdlib> // Intrinsics
# include <intrin.h> // Intrinsics
            namespace cat {
# if !defined(_M_X64)
#  define ASSEMBLY_INTEL_SYNTAX
#  define ASSEMBLY_BLOCK __asm
# endif
# define WIN32_LEAN_AND_MEAN

#elif defined(__GNUC__)
# define COMPILER_GCC /* GNU C++ Compiler */

# define PACKED __attribute__((__packed__))
# define INLINE inline
# define ASSEMBLY_ATT_SYNTAX
# define ASSEMBLY_BLOCK asm

# error "Fill in the blanks for GCC"

#elif defined(__BORLANDC__)
# define COMPILER_BORLAND /* Borland C++ Compiler */

# define INLINE __inline
# define ASSEMBLY_INTEL_SYNTAX
# define ASSEMBLY_BLOCK _asm

# error "Fill in the blanks for Borland"

#elif defined(_MSC_VER)
# define COMPILER_MSVC /* Microsoft Visual Studio C++ Compiler */

# define PACKED __declspec(align(1))
# define INLINE __inline
# define DEBUG _DEBUG
            }
# include <cstdlib> // Intrinsics
# include <intrin.h> // Intrinsics
            namespace cat {
# if !defined(_M_X64)
#  define ASSEMBLY_INTEL_SYNTAX
#  define ASSEMBLY_BLOCK __asm
# endif
# define WIN32_LEAN_AND_MEAN

#else
# error "Add your compiler to the list"
#endif


//// Architecture ////

#if defined(__sparc) || defined(__sparc__) || defined(__powerpc__) || \
	defined(__ppc__) || defined(__hppa) || defined(_MIPSEB) || defined(_POWER) || \
	defined(_M_PPC) || defined(_M_MPPC) || defined(_M_MRX000) || \
	defined(__POWERPC) || defined(m68k) || defined(powerpc) || \
	defined(sel) || defined(pyr) || defined(mc68000) || defined(is68k) || \
	defined(tahoe) || defined(ibm032) || defined(ibm370) || defined(MIPSEB) || \
	defined(__convex__) || defined(DGUX) || defined(hppa) || defined(apollo) || \
	defined(_CRAY) || defined(__hp9000) || defined(__hp9000s300) || defined(_AIX) || \
	defined(__AIX) || defined(__pyr__) || defined(hp9000s700) || defined(_IBMR2)
# define ENDIAN_BIG
#elif defined(__i386__) || defined(i386) || defined(intel) || defined(_M_IX86) || \
	  defined(__alpha__) || defined(__alpha) || defined(__ia64) || defined(__ia64__) || \
	  defined(_M_ALPHA) || defined(ns32000) || defined(__ns32000__) || defined(sequent) || \
	  defined(MIPSEL) || defined(_MIPSEL) || defined(sun386) || defined(__sun386__) || \
	  defined(__x86_64) || defined(_M_IA64) || defined(_M_X64) || defined(__bfin__)
# define ENDIAN_LITTLE
#else
# error "Add your architecture to the list"
#endif

#if defined(_WIN64)
# define ARCH_64
#else // Assuming 32-bit otherwise!
# define ARCH_32
#endif


//// Operating System ////

#if defined(__linux__)
# define OS_LINUX
#elif defined(_WIN32)
# define OS_WINDOWS
#else
# error "Add your operating system to the list"
#endif


//// Basic types ////

#if defined(COMPILER_MSVC)

	typedef unsigned __int8		u8;
	typedef signed __int8		s8;
	typedef unsigned __int16	u16;
	typedef signed __int16		s16;
	typedef unsigned __int32	u32;
	typedef signed __int32		s32;
	typedef unsigned __int64	u64;
	typedef signed __int64		s64;

#else
# error "Add your compiler's basic types"
#endif

struct u128 { u32 u[4]; };

union Float32 {
	float f;
	u32 i;

	Float32(float n) { f = n; }
	Float32(u32 n) { i = n; }
};

/*
//// Coordinate system ////

// Order of vertices in any quad
enum QuadCoords
{
	QUAD_UL, // upper left  (0,0)
	QUAD_LL, // lower left  (0,1)
	QUAD_LR, // lower right (1,1)
	QUAD_UR, // upper right (1,0)
};
*/


//// Rotation macros ////

#define ROL8(n, r)  ( ((u8)(n) << (r)) | ((u8)(n) >> ( 8 - (r))) ) /* only works for u8 */
#define ROR8(n, r)  ( ((u8)(n) >> (r)) | ((u8)(n) << ( 8 - (r))) ) /* only works for u8 */
#define ROL16(n, r) ( ((u16)(n) << (r)) | ((u16)(n) >> (16 - (r))) ) /* only works for u16 */
#define ROR16(n, r) ( ((u16)(n) >> (r)) | ((u16)(n) << (16 - (r))) ) /* only works for u16 */
#define ROL32(n, r) ( ((u32)(n) << (r)) | ((u32)(n) >> (32 - (r))) ) /* only works for u32 */
#define ROR32(n, r) ( ((u32)(n) >> (r)) | ((u32)(n) << (32 - (r))) ) /* only works for u32 */
#define ROL64(n, r) ( ((u64)(n) << (r)) | ((u64)(n) >> (64 - (r))) ) /* only works for u64 */
#define ROR64(n, r) ( ((u64)(n) >> (r)) | ((u64)(n) << (64 - (r))) ) /* only works for u64 */


//// Byte-order swapping ////

#define BOSWAP16(n) ROL16(n, 8)
#define BOSWAP32(n) ( (ROL32(n, 8) & 0x00ff00ff) | (ROL32(n, 24) & 0xff00ff00) )
#define BOSWAP64(n) ( ((u64)BOSWAP32((u32)n) << 32) | BOSWAP32((u32)(n >> 32)) )


//// Miscellaneous bitwise macros ////

#define BITCLRHI8(reg, count) ((u8)((u8)(reg) << (count)) >> (count)) /* sets to zero a number of high bits in a byte */
#define BITCLRLO8(reg, count) ((u8)((u8)(reg) >> (count)) << (count)) /* sets to zero a number of low bits in a byte */
#define BITCLRHI16(reg, count) ((u16)((u16)(reg) << (count)) >> (count)) /* sets to zero a number of high bits in a 16-bit word */
#define BITCLRLO16(reg, count) ((u16)((u16)(reg) >> (count)) << (count)) /* sets to zero a number of low bits in a 16-bit word */
#define BITCLRHI32(reg, count) ((u32)((u32)(reg) << (count)) >> (count)) /* sets to zero a number of high bits in a 32-bit word */
#define BITCLRLO32(reg, count) ((u32)((u32)(reg) >> (count)) << (count)) /* sets to zero a number of low bits in a 32-bit word */


//// Integer macros ////

#define AT_LEAST_2_BITS(n) ( (n) & ((n) - 1) )
#define LEAST_SIGNIFICANT_BIT(n) ( (n) & (u32)(-(s32)(n)) ) /* 0 -> 0 */
#define IS_POWER_OF_2(n) ( n && !AT_LEAST_2_BITS(n) )

// Bump 'n' to the next unit of 'width'
// 0=CEIL_UNIT(0, 16), 1=CEIL_UNIT(1, 16), 1=CEIL_UNIT(16, 16), 2=CEIL_UNIT(17, 16)
#define CEIL_UNIT(n, width) ( ( (n) + (width) - 1 ) / (width) )
// 0=CEIL(0, 16), 16=CEIL(1, 16), 16=CEIL(16, 16), 32=CEIL(17, 16)
#define CEIL(n, width) ( CEIL_UNIT(n, width) * (width) )


//// String and buffer macros ////

// Same as strncpy() in all ways except that the result is guaranteed to
// be a nul-terminated C string
#if defined(COMPILER_MSVC)
# define STRNCPY(dest, src, size) { strncpy_s(dest, size, src, size); (dest)[(size)-1] = '\0'; }
#else
# define STRNCPY(dest, src, size) { strncpy(dest, src, size); (dest)[(size)-1] = '\0'; }
#endif

// Because memory clearing is a frequent operation
#define MEMCLR(dest, size) memset(dest, 0, size)

// Works for arrays, also
#define OBJCLR(object) memset(&(object), 0, sizeof(object))

// Stringize
#define STRINGIZE(X) DO_STRINGIZE(X)
#define DO_STRINGIZE(X) #X


//// Intrinsics ////

#if defined(COMPILER_MSVC)

#pragma intrinsic(_rotl)
#pragma intrinsic(_rotr)
#pragma intrinsic(_rotl64)
#pragma intrinsic(_rotr64)
#pragma intrinsic(_byteswap_ushort)
#pragma intrinsic(_byteswap_ulong)
#pragma intrinsic(_byteswap_uint64)

#undef ROL32
#undef ROR32
#undef ROL64
#undef ROR64
#undef BOSWAP16
#undef BOSWAP32
#undef BOSWAP64

#define ROL32(n, r) _rotl(n, r)
#define ROR32(n, r) _rotr(n, r)
#define ROL64(n, r) _rotl64(n, r)
#define ROR64(n, r) _rotr64(n, r)
#define BOSWAP16(n) _byteswap_ushort(n)
#define BOSWAP32(n) _byteswap_ulong(n)
#define BOSWAP64(n) _byteswap_uint64(n)

#endif


// getLE() converts from little-endian word to native byte-order word 
// getBE() converts from big-endian word to native byte-order word 

#if defined(ENDIAN_LITTLE)

# define swapLE(n)
# define getLE(n) (n)
# define getLE16(n) (n)
# define getLE32(n) (n)
# define getLE64(n) (n)

	inline void swapBE(u16 &n) { n = BOSWAP16(n); }
	inline void swapBE(u32 &n) { n = BOSWAP32(n); }
	inline void swapBE(u64 &n) { n = BOSWAP64(n); }
	inline u16 getBE(u16 n) { return BOSWAP16(n); }
	inline u32 getBE(u32 n) { return BOSWAP32(n); }
	inline u64 getBE(u64 n) { return BOSWAP64(n); }
	inline u16 getBE16(u16 n) { return BOSWAP16(n); }
	inline u32 getBE32(u32 n) { return BOSWAP32(n); }
	inline u64 getBE64(u64 n) { return BOSWAP64(n); }
	inline void swapBE(s16 &n) { n = BOSWAP16((u16)n); }
	inline void swapBE(s32 &n) { n = BOSWAP32((u32)n); }
	inline void swapBE(s64 &n) { n = BOSWAP64((u64)n); }
	inline s16 getBE(s16 n) { return BOSWAP16((u16)n); }
	inline s32 getBE(s32 n) { return BOSWAP32((u32)n); }
	inline s64 getBE(s64 n) { return BOSWAP64((u64)n); }

	inline float getBE(float n) {
		Float32 c = n;
		c.i = BOSWAP32(c.i);
		return c.f;
	}

#else

# define swapBE(n)
# define getBE(n) (n)
# define getBE16(n) (n)
# define getBE32(n) (n)
# define getBE64(n) (n)

	inline void swapLE(u16 &n) { n = BOSWAP16(n); }
	inline void swapLE(u32 &n) { n = BOSWAP32(n); }
	inline void swapLE(u64 &n) { n = BOSWAP64(n); }
	inline u16 getLE(u16 n) { return BOSWAP16(n); }
	inline u32 getLE(u32 n) { return BOSWAP32(n); }
	inline u64 getLE(u64 n) { return BOSWAP64(n); }
	inline u16 getLE16(u16 n) { return BOSWAP16(n); }
	inline u32 getLE32(u32 n) { return BOSWAP32(n); }
	inline u64 getLE32(u64 n) { return BOSWAP64(n); }
	inline void swapLE(s16 &n) { n = BOSWAP16((u16)n); }
	inline void swapLE(s32 &n) { n = BOSWAP32((u32)n); }
	inline void swapLE(s64 &n) { n = BOSWAP64((u64)n); }
	inline s16 getLE(s16 n) { return BOSWAP16((u16)n); }
	inline s32 getLE(s32 n) { return BOSWAP32((u32)n); }
	inline s64 getLE(s64 n) { return BOSWAP64((u64)n); }

	inline float getLE(float n) {
		Float32 c = n;
		c.i = BOSWAP32(c.i);
		return c.f;
	}

#endif


} // namespace cat

#endif // PLATFORM_HPP
