// 06/10/09 part of common-1.0

#ifndef SHA2_HPP
#define SHA2_HPP

#include <cat/Platform.hpp>

namespace cat {


/*
	Encapsulated 256-bit message digest algorithm

	Pseudo-code for normal operation:
		DigestSHA256 sha2;
		sha2.beginDigest();	// redundant, this is called in the constructor
		sha2.performDigest(...);
		u32[8] = sha2.endDigest();

	Please note that returned pointers are only valid until the next function call.
*/
class DigestSHA256
{
	u32 work[64];
	u32 reg[8], workReg[8];
	u32 highCounter, lowCounter;

	void hashComputation();

public:
	// call beginDigest()
	DigestSHA256();

	// zero register and work memory space
	~DigestSHA256();

public:
	// initialize the 256-bit register
	void beginDigest();

	// do SHA2 rounds
	void performDigest(const void *buf, u32 len);

	// finish up by adding in the length of the input, then return the contents of the digest register
	u32 *endDigest();

	// get the previous result again
	u32 *getDigest()
	{
		return reg;
	}
};


} // namespace cat

#endif // SHA2_HPP
