// 06/10/09 part of common-1.0

#ifndef RANDOM_HPP
#define RANDOM_HPP

#include <cat/Singleton.hpp>

namespace cat {


class MersenneTwister
{
	u128 state[19937/128 + 1];
	u32 *state32;
	u32 used;

	void enforcePeriod(); // make corrections to ensure that the generator has the full period
	void round(u128 *a, u128 *b, u128 *c, u128 *d); // a = MTMIX(a,b,c,d)
	void update(); // permute the existing state into a new one

public:
	MersenneTwister();

	void init(u32 seed); // initialize the generator with a seed
	void init(u32 *seeds, u32 words); // initialize with an array of seeds
	void init(); // use system sources

	u32 generate(); // generate a 32-bit number
	void generate(void *buffer, u32 bytes); // generate a series of random numbers
};

class Random : public MersenneTwister, public Singleton<Random>
{
	SINGLETON(Random);
};


// Microsoft VC++ 7.0 stdlib srand(), rand(), old RANDU()
class StandardRand
{
protected:
	s32 seed;

public:
	StandardRand(u32 ns = 0) { seed = ns; }

	inline void srand32(u32 ns) { seed = ns; } // 32-bit version
	inline void srand16(u16 ns) { seed = ns; } // 16-bit version (yup)

	u16 rand();	// Linear Congruential Generator: X = X * M + A (mod N)
	u16 randu(); // RANDU LCG: X = X * M (mod N)
};


// Non-linear congruential 32-bit random mixing function for given x, y and seed
// This function is used in every example of Perlin noise I found online for some reason!
u32 NLCRand32(int x, int y, u32 seed);

// Uses NLCRand32 as a front-end, and then maps the output to a number between -1 and 1
float NLCRandNorm(int x, int y, u32 seed);


} // namespace cat

#endif // RANDOM_HPP
